// MainFrm.cpp : implementation of the CMainFrame class
//

#include "stdafx.h"
#include "Stage.h"
#include "math.h"
#include "MainFrm.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CMainFrame

IMPLEMENT_DYNCREATE(CMainFrame, CFrameWnd)

BEGIN_MESSAGE_MAP(CMainFrame, CFrameWnd)
	//{{AFX_MSG_MAP(CMainFrame)
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_ACTIVATEAPP()
	ON_WM_PALETTECHANGED()
	ON_WM_MOVE()
	ON_COMMAND(ID_FILE_NEW, OnFileNew)
	ON_COMMAND(ID_FILE_OPEN, OnFileOpen)
	ON_COMMAND(ID_EDIT_INSERTSHAPE, OnEditInsertShape)
	ON_COMMAND(ID_VIEW_ROTATE, OnViewRotate)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ROTATE, OnUpdateViewRotate)
	ON_COMMAND(ID_VIEW_RESET, OnViewReset)
	ON_UPDATE_COMMAND_UI(ID_VIEW_NORMALS, OnUpdateViewNormals)
	ON_COMMAND(ID_VIEW_NORMALS, OnViewNormals)
	ON_COMMAND(ID_EDIT_INSFACE, OnEditInsface)
	ON_COMMAND(ID_EDIT_BULGEFACE, OnEditBulgeface)
	ON_COMMAND(ID_EDIT_DISHFACE, OnEditDishface)
	ON_COMMAND(ID_EDIT_DEFCUBE, OnEditDefcube)
	ON_COMMAND(ID_EDIT_FLATFACECUBE, OnEditFlatfacecube)
	ON_COMMAND(ID_EDIT_INSLAND, OnEditInsland)
	ON_COMMAND(ID_EDIT_INSLAND2, OnEditInsland2)
	ON_COMMAND(ID_EDIT_SOLIDR, OnEditSolidr)
	ON_COMMAND(ID_EDIT_CONE, OnEditCone)
	ON_COMMAND(ID_EDIT_TREE, OnEditTree)
	ON_COMMAND(ID_EDIT_FOREST, OnEditForest)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

static UINT indicators[] =
{
	ID_SEPARATOR,           // status line indicator
	ID_INDICATOR_CAPS,
	ID_INDICATOR_NUM,
	ID_INDICATOR_SCRL,
};

/////////////////////////////////////////////////////////////////////////////
// CMainFrame construction/destruction

CMainFrame::CMainFrame()
{
	m_pScene = NULL;
	m_bRotate = FALSE;
	m_pShape = NULL;
}

CMainFrame::~CMainFrame()
{
}

int CMainFrame::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CFrameWnd::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	if (!m_wndToolBar.Create(this) ||
		!m_wndToolBar.LoadToolBar(IDR_MAINFRAME))
	{
		TRACE0("Failed to create toolbar\n");
		return -1;      // fail to create
	}

	if (!m_wndStatusBar.Create(this) ||
		!m_wndStatusBar.SetIndicators(indicators,
		  sizeof(indicators)/sizeof(UINT)))
	{
		TRACE0("Failed to create status bar\n");
		return -1;      // fail to create
	}

	// Create the 3D window
	if (!m_wnd3d.Create(this, IDC_3DWND)) {
		return -1;
	}

	// TODO: Remove this if you don't want tool tips or a resizeable toolbar
	m_wndToolBar.SetBarStyle(m_wndToolBar.GetBarStyle() |
		CBRS_TOOLTIPS | CBRS_FLYBY | CBRS_SIZE_DYNAMIC);

	// TODO: Delete these three lines if you don't want the toolbar to
	//  be dockable
	m_wndToolBar.EnableDocking(CBRS_ALIGN_ANY);
	EnableDocking(CBRS_ALIGN_ANY);
	DockControlBar(&m_wndToolBar);

	NewScene();
	ASSERT(m_pScene);

	// Create a shape to add
	m_pShape = new C3dShape;
	m_pShape->CreateCube(2);
	m_pScene->AddChild(m_pShape);
	Rotate(TRUE);

	return 0;
}

BOOL CMainFrame::PreCreateWindow(CREATESTRUCT& cs)
{
	// Make the initial window a fixed size
	cs.cx = 300;
	cs.cy = 350;

	return CFrameWnd::PreCreateWindow(cs);
}

BOOL CMainFrame::NewScene()
{
	Rotate(FALSE);

	// Delete any scene we might have
	if (m_pScene) {
		m_wnd3d.SetScene(NULL);
		delete m_pScene;
		m_pScene = NULL;
	}
	if (m_pShape) {
		delete m_pShape;
		m_pShape = NULL;
	}

	// Create an initial scene
	m_pScene = new C3dScene;
	if (!m_pScene->Create()) return FALSE;

	// Set up the lighting
    C3dDirLight dl;
    dl.Create(0.9, 0.9, 0.9);
    m_pScene->AddChild(&dl);
    dl.SetPosition(-2, 2, -5);
    dl.SetDirection(1, -1, 1);
    m_pScene->SetAmbientLight(0.3, 0.3, 0.3);

	m_wnd3d.SetScene(m_pScene);
	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CMainFrame diagnostics

#ifdef _DEBUG
void CMainFrame::AssertValid() const
{
	CFrameWnd::AssertValid();
}

void CMainFrame::Dump(CDumpContext& dc) const
{
	CFrameWnd::Dump(dc);
}

#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CMainFrame message handlers

void CMainFrame::RecalcLayout(BOOL bNotify) 
{
    // Rearrange the control bars and fit the 3D window in the middle
	// Let the frame rearrange the control bars
	CFrameWnd::RecalcLayout(bNotify);

	// Find the space that's left over
	CRect rc;
	RepositionBars(0,
				   0xFFFF,
				   IDC_3DWND,
				   CWnd::reposQuery,
				   &rc);
	if (IsWindow(m_wnd3d.GetSafeHwnd())) {
		m_wnd3d.MoveWindow(&rc, FALSE);
	}
}


void CMainFrame::OnDestroy() 
{
	CFrameWnd::OnDestroy();
	
	// Delete any scene we might have
	m_wnd3d.SetScene(NULL);
	if (m_pScene) {
		delete m_pScene;
		m_pScene = NULL;
	}
	if (m_pShape) {
		delete m_pShape;
		m_pShape = NULL;
	}
}

void CMainFrame::OnActivateApp(BOOL bActive, HTASK hTask) 
{
	CFrameWnd::OnActivateApp(bActive, hTask);
	
	// Tell the 3D window about the new state
	m_wnd3d.SendMessage(WM_ACTIVATEAPP,
						(WPARAM)bActive,
						(LPARAM)hTask);
}

void CMainFrame::OnPaletteChanged(CWnd* pFocusWnd) 
{
	// Let the 3D window know
	m_wnd3d.SendMessage(WM_PALETTECHANGED,
					   pFocusWnd ? (WPARAM)pFocusWnd->GetSafeHwnd() : 0);
}

void CMainFrame::OnMove(int x, int y) 
{
	CFrameWnd::OnMove(x, y);
	
	// Let the 3D window know the frame has moved
	m_wnd3d.SendMessage(WM_MOVE, 0, MAKELPARAM(0, 0));
}

void CMainFrame::OnFileNew() 
{
	NewScene();
}

BOOL CMainFrame::OpenFile(const char* pszPath)
{
	// Try for a shape file
	C3dShape sh;
	const char* pszFile = sh.Load(pszPath);
	if (!pszFile) return FALSE;

	// Create a new scene
	NewScene();
	ASSERT(m_pScene);

	// Attach the new shape to the scene
	m_pScene->AddChild(&sh);

	// Add the name to the recent file list
	AfxGetApp()->AddToRecentFileList(pszFile);
	return TRUE;
}

void CMainFrame::OnFileOpen() 
{
	OpenFile(NULL);
}

void CMainFrame::OnEditInsertShape() 
{
	CShapeDlg dlg("Insert Shape", this);

	if (dlg.DoModal() != IDOK) return;

	NewScene();
	if (dlg.m_pShape) {
		ASSERT(m_pScene);
		m_pShape = dlg.m_pShape;
		m_pScene->AddChild(m_pShape);
	}
}

void CMainFrame::Rotate(BOOL bRotate)
{
	if (bRotate == m_bRotate) return;
	if (!m_pShape) return;
	if (bRotate) {
		m_pShape->SetRotation(1, 1, 1, 0.02);
	} else {
		m_pShape->SetRotation(1, 1, 1, 0);
	}
	m_bRotate = bRotate;
}

void CMainFrame::OnViewRotate() 
{
	Rotate(!m_bRotate);
}

void CMainFrame::OnUpdateViewRotate(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_pShape != NULL);
	pCmdUI->SetCheck(m_bRotate);	
}

void CMainFrame::OnViewReset() 
{
	Rotate(FALSE);
	if (m_pShape) {
		m_pShape->SetDirection(0, 0, 1, 0, 1, 0);
	}
}

void CMainFrame::OnUpdateViewNormals(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_pShape != NULL);	
}

void CMainFrame::OnViewNormals() 
{
	ASSERT(m_pShape);

	// Get the source mesh builder
	IDirect3DRMMeshBuilder* pMB = m_pShape->GetMeshBuilder();
	ASSERT(pMB);
	
	// get the face list
	IDirect3DRMFaceArray* pIFA = NULL;
	HRESULT hr = pMB->GetFaces(&pIFA);
	ASSERT(SUCCEEDED(hr));

	// Create the pointer shapes
	C3dShape rod;
	rod.CreateRod(0, 0, 0, 0, 0, 0.5, 0.1);
	C3dShape cone;
	cone.CreateCone(0, 0, 0.5, 0.2, TRUE, 0, 0, 1, 0, FALSE);

	// walk the face list
	ULONG nFaces = pMB->GetFaceCount();
	for (ULONG i = 0; i < nFaces; i++) {

		IDirect3DRMFace* pIFace = NULL;
		hr = pIFA->GetElement(i, &pIFace);
		ASSERT(SUCCEEDED(hr));

		// get the vertex count
		int iVert = pIFace->GetVertexCount();

		// walk the vertex list
		for (int iv = 0; iv < iVert; iv++) {
			C3dVector vPos;
			C3dVector vNorm;
			hr = pIFace->GetVertex(iv, &vPos, &vNorm);
			ASSERT(SUCCEEDED(hr));

			// Make sure the normal is valid
			if (!vNorm.IsNull()) {

				// Add a pointer at this vertex in the direction
				// of the normal
				C3dShape* pr = rod.Clone();
				C3dShape* pc = cone.Clone();
				pr->AddChild(pc);
				m_pShape->AddChild(pr);
				pr->SetPosition(vPos);
				pr->SetDirection(vNorm);
				delete pr;
				delete pc;
			}
		}
		pIFace->Release();
	}
	pIFA->Release();
}

void CMainFrame::OnEditInsface() 
{
	// Insert a simple flat face, and
	// define list of vertices
	D3DVECTOR vlist [] = {
		{-2, -2, -2},
		{-2, -2,  2},
		{ 2, -2,  2},
		{ 2, -2, -2}
	};

	// Get the number of vertices in array
	int nv = sizeof(vlist) / sizeof(D3DVECTOR);

	// Define the vertices used in each face as
	// count, vert1, vert2, and so on
	int flist [] = {4, 0, 1, 2, 3, // 4 Vertices, ...
					0 // End of the face data list
	};

	// Put this shape into a new scene
	NewScene();

	// Create shape from list of vertices and
	// face data list
	m_pShape = new C3dShape;
	m_pShape->Create(vlist, nv, flist);

	// Add the shape to the scene
	m_pScene->AddChild(m_pShape);
}

void CMainFrame::OnEditBulgeface() 
{
	// A bulging face
	D3DVECTOR vlist [] = {
		{-2, -2, -2},
		{-2, -2,  2},
		{ 2, -2,  2},
		{ 2, -2, -2}
	};
	int nv = sizeof(vlist) / sizeof(D3DVECTOR);

	D3DVECTOR nlist [] = {
		{-1,  1, -1},
		{-1,  1,  1},
		{ 1,  1,  1},
		{ 1,  1, -1}
	};
	int nn = sizeof(nlist) / sizeof(D3DVECTOR);

	int flist [] = {4, 0, 0, 1, 1, 2, 2, 3, 3,
					0
	};

	NewScene();
	m_pShape = new C3dShape;
	m_pShape->Create(vlist, nv, nlist, nn, flist);
	m_pScene->AddChild(m_pShape);
}

void CMainFrame::OnEditDishface() 
{
	// A dished face
	D3DVECTOR vlist [] = {
		{-2, -2, -2},
		{-2, -2,  2},
		{ 2, -2,  2},
		{ 2, -2, -2}
	};
	int nv = sizeof(vlist) / sizeof(D3DVECTOR);

	D3DVECTOR nlist [] = {
		{ 1,  1,  1},
		{ 1,  1, -1},
		{-1,  1, -1},
		{-1,  1,  1}
	};
	int nn = sizeof(nlist) / sizeof(D3DVECTOR);

	int flist [] = {4, 0, 0, 1, 1, 2, 2, 3, 3,
					0
	};

	NewScene();
	m_pShape = new C3dShape;
	m_pShape->Create(vlist, nv, nlist, nn, flist);
	m_pScene->AddChild(m_pShape);
}

void CMainFrame::OnEditDefcube() 
{
    D3DVECTOR vlist [] = {
        {-1, -1, -1},
        {-1, -1,  1},
        { 1, -1,  1},
        { 1, -1, -1},
        {-1,  1, -1},
        {-1,  1,  1},
        { 1,  1,  1},
        { 1,  1, -1}
    };
	int nv = sizeof(vlist) / sizeof(D3DVECTOR);

	int flist [] = {4, 0, 3, 2, 1,
                    4, 3, 7, 6, 2,
                    4, 4, 5, 6, 7,
                    4, 0, 1, 5, 4,
                    4, 0, 4, 7, 3,
                    4, 2, 6, 5, 1,
                    0
    };

	NewScene();
	m_pShape = new C3dShape;
	m_pShape->Create(vlist, nv, flist);
	m_pScene->AddChild(m_pShape);
}

void CMainFrame::OnEditFlatfacecube() 
{
    D3DVECTOR vlist [] = {
        {-1, -1, -1},
        {-1, -1,  1},
        { 1, -1,  1},
        { 1, -1, -1},
        {-1,  1, -1},
        {-1,  1,  1},
        { 1,  1,  1},
        { 1,  1, -1}
    };
	int nv = sizeof(vlist) / sizeof(D3DVECTOR);

	D3DVECTOR nlist [] = {
		{ 1,  0,  0},
		{ 0,  1,  0},
		{ 0,  0,  1},
		{-1,  0,  0},
		{ 0, -1,  0},
		{ 0,  0, -1}
	};
	int nn = sizeof(nlist) / sizeof(D3DVECTOR);
    
	int flist [] = {4, 0, 4, 3, 4, 2, 4, 1, 4,
                    4, 3, 0, 7, 0, 6, 0, 2,	0,
                    4, 4, 1, 5, 1, 6, 1, 7,	1,
                    4, 0, 3, 1, 3, 5, 3, 4,	3,
                    4, 0, 5, 4, 5, 7, 5, 3,	5,
                    4, 2, 2, 6, 2, 5, 2, 1,	2,
                    0
    };

	NewScene();
	m_pShape = new C3dShape;
	m_pShape->Create(vlist, nv, nlist, nn, flist);
	m_pScene->AddChild(m_pShape);
}

double LandscapeFn(double x, double z, void* pArg)
{
	return -2.0 + (double)(rand() % 100) / 100;
}

// Create a shape to represent landscape
void CMainFrame::OnEditInsland() 
{
	// Create a shape using a height function
	NewScene();
	m_pShape = new C3dShape;
	m_pShape->CreateSurface(-5, 5, 1, -10, 10, 1,
		                    LandscapeFn, NULL);
	m_pScene->AddChild(m_pShape);
}

// Create a landscape by subdividing faces
void CMainFrame::OnEditInsland2() 
{
	// Create the initial shape, which is a simple
	// rectangular face.
	// Insert a simple flat face.
	double x = 10;
	double z1 = -10;
	double z2 = 20;
	D3DVECTOR vlist [] = {
		{-x, -4, z1},
		{-x, -4, z2},
		{ x, -4, z2},
		{ x, -4, z1}
	};
	int nv = sizeof(vlist) / sizeof(D3DVECTOR);

	int flist [] = {4, 0, 1, 2, 3,
					0
	};

	NewScene();
	m_pShape = new C3dShape;
	m_pShape->Create(vlist, nv, flist);

	// Progressivly subdivide faces 
	int iCycles = 5;
	double dHeight = 1.0;
	while (iCycles--) {

		// Get current list of faces
		int nFaces = m_pShape->GetFaceCount();
		IDirect3DRMMeshBuilder* pMB =
			                    m_pShape->GetMeshBuilder();
		ASSERT(pMB);
		IDirect3DRMFaceArray* pIFA = NULL;
		HRESULT hr;
		hr = pMB->GetFaces(&pIFA);
		ASSERT(SUCCEEDED(hr));

		// Create new shape to add new faces to
		C3dShape* pNewShape = new C3dShape;

		// Walk through the face list
		for (int iFace = 0; iFace < nFaces; iFace++) {

			IDirect3DRMFace* pIFace = NULL;
			hr = pIFA->GetElement(iFace, &pIFace);
			ASSERT(SUCCEEDED(hr));
			ASSERT(pIFace);

			// Get vertex count
			DWORD nVert = pIFace->GetVertexCount();

			// Allocate buffers
			D3DVECTOR* pVert = new D3DVECTOR [nVert];;

			// Get the vertex data
			hr = pIFace->GetVertices(&nVert,
									 pVert,
									 NULL);
			ASSERT(SUCCEEDED(hr));
			ASSERT(pVert);
			ASSERT(nVert > 2);

			// Allocate space for the new vertex and face data 
			// lists
			D3DVECTOR* NewVert = new D3DVECTOR [nVert + 1];
			int* NewFaceData = new int [4 * nVert + 1];

			// Copy old vertices and generate coordinate
			// sums
			C3dVector vNew(0, 0, 0);
			for (DWORD i = 0; i < nVert; i++) {
				NewVert[i] = pVert[i];
				vNew.x += pVert[i].x;
				vNew.y += pVert[i].y;
				vNew.z += pVert[i].z;
			}

			// Compute new vertex in plane of face
			vNew.x /= nVert;
			vNew.y /= nVert;
			vNew.z /= nVert;

			// Add a random height factor
			double dh = dHeight * (1.0 - ((double)(rand() % 
				        100)) / 50.0);
			vNew.y += dh;

			// Add new vertex
			NewVert[nVert] = vNew;

			// Create face data
			int *pfd = NewFaceData;
			for (i = 0; i < nVert; i++) {
				*pfd++ = 3;
				*pfd++ = i;
				*pfd++ = (i+1) % nVert;
				*pfd++ = nVert; // The new vertex
			}
			*pfd = 0;

			// Add new faces to shape
			pNewShape->AddFaces(NewVert, nVert+1, NULL, 0, 
				                NewFaceData);

			// Free the vector and face storage
			delete [] NewVert;
			delete [] NewFaceData;

			// Free the vertex data storage
			delete [] pVert;

			// Release the face
			pIFace->Release();
		}

		// Release the face array
		pIFA->Release();

		pMB->GenerateNormals();
		
		// Note: Don't release the mesh builder interface.

		// Delete old shape and make new one current.
		delete m_pShape;
		m_pShape = pNewShape;
	}

	// Add final shape to scene
	m_pScene->AddChild(m_pShape);
}

double RevolveFn(double z, void* pArg)
{
	if (z < -1.1) {
		return sqrt(1 - (z + 2)*(z + 2));
	} else if (z > 0.8) {
		return sqrt(2 - (z - 2)*(z - 2));
	} else {
		return 0.5;
	}
}

// Create a solid of revolution
void CMainFrame::OnEditSolidr() 
{
	// Create a shape using a height function
	NewScene();
	m_pShape = new C3dShape;
	m_pShape->CreateRSolid(-3.0, 2.2, 0.2, TRUE, TRUE,
		                   RevolveFn, NULL, 16);
	m_pScene->AddChild(m_pShape);

	// Turn it to see its side view
	m_pShape->SetDirection(0, -1, 0);
}

// Create a cone
void CMainFrame::OnEditCone() 
{
	NewScene();
	m_pShape = new C3dShape;
	m_pShape->CreateCone(0, -2, 0, 2, TRUE, 3, 3, 3, 0, FALSE, 16);
	m_pScene->AddChild(m_pShape);
}

void CMainFrame::OnEditTree() 
{
	NewScene();
	m_pShape = new C3dShape;
	m_pShape->CreateCone(0, 0, 0, 1, TRUE, 0, 4, 0, 0, FALSE);
	m_pScene->AddChild(m_pShape);
	C3dShape trunk;
	trunk.CreateRod(0, -2, 0, 0, 0, 0, 0.2);
	m_pShape->AddChild(&trunk);
}

void CMainFrame::OnEditForest() 
{
	NewScene();

	// Create the first tree and its trunk
	m_pShape = new C3dShape;
	m_pShape->CreateCone(0, 0, 0, 1, TRUE, 0, 4, 0, 0,
		                 FALSE);
	C3dShape trunk;
	trunk.CreateRod(0, -2, 0, 0, 0, 0, 0.2);
	m_pShape->AddChild(&trunk);

	// Add tree to scene
	m_pScene->AddChild(m_pShape);

	// Get position and orientation of tree trunk
	// relative to its parent, the tree
	C3dVector p, d, u;
	trunk.GetPosition(p);
	trunk.GetDirection(d, u);

	// Clone tree 99 times
	for (int i = 0; i < 99; i++) {

		// Clone tree and trunk
		C3dShape* pTree = m_pShape->Clone();
		C3dShape* pTrunk = trunk.Clone();
		pTree->AddChild(pTrunk);

		// Set relative position of the trunk
		// to the tree
		pTrunk->SetPosition(p);
		pTrunk->SetDirection(d, u);

		// Add clones as children of first tree
		// so we can rotate entire forest
		m_pShape->AddChild(pTree);

		// Place new tree in scene
		pTree->SetPosition(((double)(rand() % 100) / 5) - 10.0,
						   0,
						   (double)(rand() % 100) / 5,
						   m_pScene);

		// Delte containers
		delete pTrunk;
		delete pTree;
	}
}
